import torch
from torch import nn


class Hierarchical1d(nn.Module):
    """
    Splits the incoming sequence into two sequences by parity index.

    Returns
    -------
    torch.Tensor, even, odd
    """

    def __init__(self):
        super(Hierarchical1d, self).__init__()

    @staticmethod
    def odd(x):
        return x[:, ::2]

    @staticmethod
    def even(x):
        return x[:, 1::2]

    def forward(self, x):
        """Returns the odd and even part"""
        return self.odd(x), self.even(x)


class MoveAvg(nn.Module):
    """Take the moving average of the input sequence.
            accept a 2-dimensional tensor x, and return a 2-dimensional tensor of the shape:
                (x.shape[0], x.shape[1] - kernel_size)

    Parameters
    ----------
    kernel_size : int, moving average window size
    stride : int, the jump-step-length in calculating the moving average
    padding : str, only support to 'same' and 'valid', valid means not to pad, same means padding with
    the nearest valid value to the null value

    Returns
    -------
    torch.Tensor
    """
    def __init__(self, kernel_size, stride=1):
        super(MoveAvg, self).__init__()
        self.kernel_size, self.stride = kernel_size, stride
        # (, , x.shape[-1] - kernel_size + 1 - stride)
        self.moving_layer = nn.AvgPool1d(kernel_size, stride=stride, padding=0)

    def forward(self, x):
        moving_mean = self.moving_layer(x)
        return torch.concat((moving_mean[:, 0].reshape((-1, 1)), moving_mean), dim=1)


class DimensionConv1d(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, **kwargs):
        super(DimensionConv1d, self).__init__()
        self.conv1d = nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            **kwargs
        )

    def forward(self, x):
        assert x.ndim == 2
        return self.conv1d(x.view(x.shape[0], x.shape[1], 1)).squeeze()


class DifferentialLayer(nn.Module):
    """Differencing the incoming tensor along an axis.
            accept a 2-dimensional tensor x, and return a 2-dimensional tensor of the shape

    Parameters
    ----------
    axis: int, default to -1
    diff_n: int, default to 1

    Returns
    -------
    torch.Tensor
    """
    def __init__(self, axis=-1, diff_n=1):
        super(DifferentialLayer, self).__init__()
        self.diff_n = diff_n
        self.axis = axis

    def forward(self, x):
        assert x.ndim == 2
        res = torch.diff(x, n=self.diff_n, dim=self.axis)
        return res



import numpy as np, copy
import pandas as pd
from scipy import stats
from joblib import Parallel, delayed
from sklearn.linear_model import LinearRegression
from sklearn.base import TransformerMixin

from spinesTS.utils import check_is_fitted
from spinesTS.preprocessing import lag_splits
from spinesTS.base import TableFeatureGenerateMixin


class ContinuousFeatureGenerator:
    """Extract features for two dims series.

    Parameters
    ----------
    window_size : int or float. The length of the interval around the outlier,
            if float-type, represents the ratio of the range of each move to the length of each data sample;
            if integer-type, represents the step size of each move.
    drop_init_features : bool, whether to drop the initial features

    Returns
    -------
    None
    """

    def __init__(
            self,
            drop_init_features=False,
            columns_prefix='continuous_feature_'
    ):
        """
        Extract features for continuous sequences of inputs.
        """

        self._drop_init_features = drop_init_features
        self.__spinesTS_is_fitted__ = False

        self.columns_prefix = columns_prefix
        self.columns = []

    def get_usual_statistical(self, x):
        """Get descriptive statistics.

        Parameters
        ----------
        x : array-like

        Returns
        -------
        numpy.ndarray
        """
        _mean = np.mean(x, axis=-1).reshape(-1, 1)
        _kurt = stats.kurtosis(x, axis=-1, nan_policy='omit').reshape(-1, 1)
        _skew = stats.skew(x, axis=-1, nan_policy='omit').reshape(-1, 1)
        _min = np.min(x, axis=-1).reshape(-1, 1)
        _max = np.max(x, axis=-1).reshape(-1, 1)
        _median = np.median(x, axis=-1).reshape(-1, 1)
        _var = np.var(x, axis=-1).reshape(-1, 1)

        return np.concatenate((_mean, _kurt, _skew, _min, _max, _median, _var), axis=-1), [
            self.columns_prefix + 'mean', self.columns_prefix + 'kurt', self.columns_prefix + 'skew',
            self.columns_prefix + 'min', self.columns_prefix + 'max', self.columns_prefix + 'median',
            self.columns_prefix + 'var'
        ]

    def get_linearity(self, x):
        """Get the linear fitting feature.

        Parameters
        ----------
        x : array-like

        Returns
        -------
        numpy.ndarray
        """

        def fit_each_x(_, name='coef'):
            reg = LinearRegression(n_jobs=-1)
            reg.fit(np.reshape(range(len(_)), (-1, 1)), _)

            return list(reg.coef_)

        _coeff = Parallel(n_jobs=-1)(delayed(fit_each_x)(_) for _ in x)

        return np.array(_coeff), [self.columns_prefix + 'coeff']

    def get_entropy(self, x):
        """Get the cross entropy features.

        Parameters
        -----------
        x : array-like

        Returns
        -------
        numpy.ndarray
        """
        _ = Parallel(n_jobs=-1)(delayed(stats.entropy)(_) for _ in x)
        _ = np.array(_).reshape((-1, 1))
        res = np.where((np.isinf(_) | np.isnan(_)), 0., _)

        return res, [
            self.columns_prefix + 'entropy'
        ]

    def fit(self, x):
        """Fit the inputting matrix.

        Parameters
        -----------
        x : array-like

        Returns
        -------
        self
        """

        assert isinstance(x, np.ndarray) and x.ndim == 2
        self.__spinesTS_is_fitted__ = True
        return self

    def fit_transform(self, x):
        """
        Fit and extract the features of the inputting matrix.

        Parameters
        -----------
        x : array-like
        inplace : Whether to transform x in place or to return a copy.

        Returns
        -------
        numpy.ndarray
        """
        check_is_fitted(self)
        self.fit(x)

        return self.transform(x)

    def transform(self, X):
        """Extract the features of the inputting matrix.

        Parameters
        -----------
        X : array-like

        Returns
        -------
        numpy.ndarray
        """
        check_is_fitted(self)
        x = copy.deepcopy(X)

        gus = self.get_usual_statistical(x)
        ge = self.get_entropy(x)
        gl = self.get_linearity(x)
        res = [gus, ge, gl]
        for c in res:
            self.columns.extend(c[1])

        if self._drop_init_features:
            return np.concatenate([i[0] for i in res], axis=-1)
        else:
            return np.concatenate([x, *[i[0] for i in res]], axis=-1)


class DataExtendFeatures(TableFeatureGenerateMixin, TransformerMixin):
    def __init__(self, date_col, drop_date_col=True, format=None, columns_prefix='timefeatures_'):
        self.date_col = date_col
        self.drop_date_col = drop_date_col
        self.format = format
        if not isinstance(columns_prefix, str):
            columns_prefix = 'timefeatures_'

        self.columns_prefix = columns_prefix

        self.__spinesTS_is_fitted__ = False

    @staticmethod
    def day_of_quarter(s):
        if 1 <= s.month <= 3:
            return s.dayofyear
        elif 4 <= s.month <= 6:
            return s.dayofyear - (
                    pd.to_datetime(str(s.year) + '-03-31') - pd.to_datetime(str(s.year) + '-01-01')).days - 1
        elif 7 <= s.month <= 9:
            return s.dayofyear - (
                    pd.to_datetime(str(s.year) + '-06-30') - pd.to_datetime(str(s.year) + '-01-01')).days - 1
        else:
            return s.dayofyear - (
                    pd.to_datetime(str(s.year) + '-9-30') - pd.to_datetime(str(s.year) + '-01-01')).days - 1

    @staticmethod
    def day2season(s, season, year_lag=0):
        if season == 'spring':
            time_point = pd.to_datetime(str(s.year - year_lag) + '-03-01')
        elif season == 'summer':
            time_point = pd.to_datetime(str(s.year - year_lag) + '-06-01')
        elif season == 'autumn':
            time_point = pd.to_datetime(str(s.year - year_lag) + '-09-01')
        else:
            time_point = pd.to_datetime(str(s.year - year_lag) + '-12-01')

        return (s - time_point).days

    @staticmethod
    def season(s):
        """0: Winter - 1: Spring - 2: Summer - 3: Fall"""
        if s.month in [12, 1, 2]:
            return 0
        elif s.month in [3, 4, 5]:
            return 1
        elif s.month in [6, 7, 8]:
            return 2
        else:
            return 3

    def fit(self, df):
        self.__spinesTS_is_fitted__ = True

        return self

    def transform(self, df):
        """Date features generation"""
        check_is_fitted(self)

        assert isinstance(df, pd.DataFrame)
        x = df[self.date_col].copy().to_frame()
        ds_col = pd.to_datetime(x[self.date_col], format=self.format)

        # Basic timestamp features
        x[self.columns_prefix + 'hour'] = ds_col.dt.hour
        x[self.columns_prefix + 'minute'] = ds_col.dt.minute
        x[self.columns_prefix + 'weekday'] = ds_col.dt.weekday
        x[self.columns_prefix + 'week'] = ds_col.apply(lambda s: s.week)
        x[self.columns_prefix + 'month'] = ds_col.dt.month
        x[self.columns_prefix + 'quarter'] = ds_col.dt.quarter
        x[self.columns_prefix + 'day_of_month'] = ds_col.dt.day
        x[self.columns_prefix + 'day_of_year'] = ds_col.dt.dayofyear
        x[self.columns_prefix + 'day_of_quarter'] = ds_col.apply(lambda s: self.day_of_quarter(s))
        x[self.columns_prefix + 'week_of_month'] = ds_col.apply(lambda d: (d.day - 1) // 7 + 1)
        x[self.columns_prefix + "season"] = ds_col.apply(lambda s: self.season(s))

        # Boolean Flag feature
        x[self.columns_prefix + 'is_weekend'] = ds_col.apply(lambda s: 0 if s.weekday() in [5, 6] else 1)
        x[self.columns_prefix + 'is_start_of_week'] = ds_col.apply(lambda s: 0 if s.weekday() != 0 else 1)
        x[self.columns_prefix + 'is_end_of_week'] = ds_col.apply(lambda s: 0 if s.weekday() != 4 else 1)
        x[self.columns_prefix + 'is_start_of_month'] = ds_col.dt.is_month_start
        x[self.columns_prefix + 'is_middle_of_month'] = ds_col.apply(lambda s: 0 if s.day != 15 else 1)
        x[self.columns_prefix + 'is_end_of_month'] = ds_col.dt.is_month_end
        x[self.columns_prefix + 'is_quarter_start'] = ds_col.dt.is_quarter_start
        x[self.columns_prefix + 'is_year_start'] = ds_col.dt.is_year_start
        x[self.columns_prefix + 'is_year_end'] = ds_col.dt.is_year_end

        # Time difference feature
        x[self.columns_prefix + 'day_to_mid_quarter'] = x[self.columns_prefix + 'day_of_quarter'] - 15
        x[self.columns_prefix + 'day_to_start_quarter'] = x[self.columns_prefix + 'day_of_quarter'] - 1
        x[self.columns_prefix + 'day_to_end_quarter'] = x[self.columns_prefix + 'day_of_quarter'] - 90
        x[self.columns_prefix + 'day_to_monday'] = abs(ds_col.dt.weekday - 1)  # 仅考虑当周周一
        x[self.columns_prefix + 'day_to_friday'] = abs(ds_col.dt.weekday - 5)  # 仅考虑当周周五
        x[self.columns_prefix + 'day_to_middle_week'] = abs(ds_col.dt.weekday - 3)  # 仅考虑当周周三
        x[self.columns_prefix + 'day_to_start_timestamp'] = (ds_col - ds_col.min()).dt.days
        x[self.columns_prefix + 'hour_to_start_timestamp'] = (ds_col - ds_col.min()).dt.days * 24
        x[self.columns_prefix + 'minute_to_start_timestamp'] = (ds_col - ds_col.min()).dt.days * 24 * 60
        x[self.columns_prefix + 'sec_to_start_timestamp'] = (ds_col - ds_col.min()).dt.days * 24 * 3600
        x[self.columns_prefix + 'month_to_start_timestamp'] = (ds_col - ds_col.min()).dt.days / 30
        x[self.columns_prefix + 'week_to_start_timestamp'] = (ds_col - ds_col.min()).dt.days / 7
        x[self.columns_prefix + 'quarter_to_start_timestamp'] = (ds_col - ds_col.min()).dt.days / 90
        x[self.columns_prefix + 'year_to_start_timestamp'] = (ds_col - ds_col.min()).dt.days / 365

        # season Time difference feature
        x[self.columns_prefix + 'day_to_spring'] = ds_col.apply(lambda s: self.day2season(s, 'spring'))
        x[self.columns_prefix + 'day_to_summer'] = ds_col.apply(lambda s: self.day2season(s, 'summer'))
        x[self.columns_prefix + 'day_to_autumn'] = ds_col.apply(lambda s: self.day2season(s, 'autumn'))
        x[self.columns_prefix + 'day_to_winter'] = ds_col.apply(lambda s: self.day2season(s, 'winter'))
        x[self.columns_prefix + 'week_to_spring'] = ds_col.apply(lambda s: self.day2season(s, 'spring')) / 7
        x[self.columns_prefix + 'week_to_summer'] = ds_col.apply(lambda s: self.day2season(s, 'summer')) / 7
        x[self.columns_prefix + 'week_to_autumn'] = ds_col.apply(lambda s: self.day2season(s, 'autumn')) / 7
        x[self.columns_prefix + 'week_to_winter'] = ds_col.apply(lambda s: self.day2season(s, 'winter')) / 7
        x[self.columns_prefix + 'month_to_spring'] = ds_col.apply(lambda s: self.day2season(s, 'spring')) / 30
        x[self.columns_prefix + 'month_to_summer'] = ds_col.apply(lambda s: self.day2season(s, 'summer')) / 30
        x[self.columns_prefix + 'month_to_autumn'] = ds_col.apply(lambda s: self.day2season(s, 'autumn')) / 30
        x[self.columns_prefix + 'month_to_winter'] = ds_col.apply(lambda s: self.day2season(s, 'winter')) / 30
        x[self.columns_prefix + 'quarter_to_spring'] = ds_col.apply(lambda s: self.day2season(s, 'spring')) / 90
        x[self.columns_prefix + 'quarter_to_summer'] = ds_col.apply(lambda s: self.day2season(s, 'summer')) / 90
        x[self.columns_prefix + 'quarter_to_autumn'] = ds_col.apply(lambda s: self.day2season(s, 'autumn')) / 90
        x[self.columns_prefix + 'quarter_to_winter'] = ds_col.apply(lambda s: self.day2season(s, 'winter')) / 90

        x[self.columns_prefix + 'day_to_next_year_spring'] = ds_col.apply(lambda s: self.day2season(s, 'spring', -1))
        x[self.columns_prefix + 'day_to_next_year_summer'] = ds_col.apply(lambda s: self.day2season(s, 'summer', -1))
        x[self.columns_prefix + 'day_to_next_year_autumn'] = ds_col.apply(lambda s: self.day2season(s, 'autumn', -1))
        x[self.columns_prefix + 'day_to_next_year_winter'] = ds_col.apply(lambda s: self.day2season(s, 'winter', -1))
        x[self.columns_prefix + 'week_to_next_year_spring'] = ds_col.apply(
            lambda s: self.day2season(s, 'spring', -1)) / 7
        x[self.columns_prefix + 'week_to_next_year_summer'] = ds_col.apply(
            lambda s: self.day2season(s, 'summer', -1)) / 7
        x[self.columns_prefix + 'week_to_next_year_autumn'] = ds_col.apply(
            lambda s: self.day2season(s, 'autumn', -1)) / 7
        x[self.columns_prefix + 'week_to_next_year_winter'] = ds_col.apply(
            lambda s: self.day2season(s, 'winter', -1)) / 7
        x[self.columns_prefix + 'month_to_next_year_spring'] = ds_col.apply(
            lambda s: self.day2season(s, 'spring', -1)) / 30
        x[self.columns_prefix + 'month_to_next_year_summer'] = ds_col.apply(
            lambda s: self.day2season(s, 'summer', -1)) / 30
        x[self.columns_prefix + 'month_to_next_year_autumn'] = ds_col.apply(
            lambda s: self.day2season(s, 'autumn', -1)) / 30
        x[self.columns_prefix + 'month_to_next_year_winter'] = ds_col.apply(
            lambda s: self.day2season(s, 'winter', -1)) / 30
        x[self.columns_prefix + 'quarter_to_next_year_spring'] = ds_col.apply(
            lambda s: self.day2season(s, 'spring', -1)) / 90
        x[self.columns_prefix + 'quarter_to_next_year_summer'] = ds_col.apply(
            lambda s: self.day2season(s, 'summer', -1)) / 90
        x[self.columns_prefix + 'quarter_to_next_year_autumn'] = ds_col.apply(
            lambda s: self.day2season(s, 'autumn', -1)) / 90
        x[self.columns_prefix + 'quarter_to_next_year_winter'] = ds_col.apply(
            lambda s: self.day2season(s, 'winter', -1)) / 90

        x[self.columns_prefix + 'day_to_last_year_spring'] = ds_col.apply(lambda s: self.day2season(s, 'spring', 1))
        x[self.columns_prefix + 'day_to_last_year_summer'] = ds_col.apply(lambda s: self.day2season(s, 'summer', 1))
        x[self.columns_prefix + 'day_to_last_year_autumn'] = ds_col.apply(lambda s: self.day2season(s, 'autumn', 1))
        x[self.columns_prefix + 'day_to_last_year_winter'] = ds_col.apply(lambda s: self.day2season(s, 'winter', 1))
        x[self.columns_prefix + 'week_to_last_year_spring'] = ds_col.apply(
            lambda s: self.day2season(s, 'spring', 1)) / 7
        x[self.columns_prefix + 'week_to_last_year_summer'] = ds_col.apply(
            lambda s: self.day2season(s, 'summer', 1)) / 7
        x[self.columns_prefix + 'week_to_last_year_autumn'] = ds_col.apply(
            lambda s: self.day2season(s, 'autumn', 1)) / 7
        x[self.columns_prefix + 'week_to_last_year_winter'] = ds_col.apply(
            lambda s: self.day2season(s, 'winter', 1)) / 7
        x[self.columns_prefix + 'month_to_last_year_spring'] = ds_col.apply(
            lambda s: self.day2season(s, 'spring', 1)) / 30
        x[self.columns_prefix + 'month_to_last_year_summer'] = ds_col.apply(
            lambda s: self.day2season(s, 'summer', 1)) / 30
        x[self.columns_prefix + 'month_to_last_year_autumn'] = ds_col.apply(
            lambda s: self.day2season(s, 'autumn', 1)) / 30
        x[self.columns_prefix + 'month_to_last_year_winter'] = ds_col.apply(
            lambda s: self.day2season(s, 'winter', 1)) / 30
        x[self.columns_prefix + 'quarter_to_last_year_spring'] = ds_col.apply(
            lambda s: self.day2season(s, 'spring', 1)) / 90
        x[self.columns_prefix + 'quarter_to_last_year_summer'] = ds_col.apply(
            lambda s: self.day2season(s, 'summer', 1)) / 90
        x[self.columns_prefix + 'quarter_to_last_year_autumn'] = ds_col.apply(
            lambda s: self.day2season(s, 'autumn', 1)) / 90
        x[self.columns_prefix + 'quarter_to_last_year_winter'] = ds_col.apply(
            lambda s: self.day2season(s, 'winter', 1)) / 90

        df = pd.concat((df, x), axis=1)
        return df if not self.drop_date_col else df.drop(columns=self.date_col)


class DailyTargetExtendFeatures(TableFeatureGenerateMixin, TransformerMixin):
    def __init__(self, target_col, date_col, lag, columns_prefix='targetfeatures_'):
        self.target_col = target_col
        self.date_col = date_col
        self.lag = lag

        if not isinstance(columns_prefix, str):
            columns_prefix = 'targetfeatures_'

        self.columns_prefix = columns_prefix
        self.__spinesTS_is_fitted__ = False

    def fit(self, df):
        self.__spinesTS_is_fitted__ = True

        return self

    def weekday_groupby(self, df):
        x = df[[self.date_col, self.target_col]].copy()

        x['weekday'] = x[self.date_col].dt.weekday

        days = {
            'quarter': 90,
            'month': 30,
            'double_week': 14,
        }

        for col in days:
            d_length = days[col]
            mean_res = []
            median_res = []
            p25 = []
            p75 = []

            for i in range(x.shape[0]):
                if i - d_length <= 0:
                    start = 0
                else:
                    start = i - d_length

                if i <= 7:
                    mean_res.append(0)
                    median_res.append(0)
                    p25.append(0)
                    p75.append(0)
                else:
                    weekday = x['weekday'].iloc[i]
                    ts = x.iloc[start: i].query(f"weekday == {weekday}")[self.target_col]
                    mean_res.append(ts.mean())
                    median_res.append(ts.median())
                    p25.append(ts.quantile(q=0.25))
                    p75.append(ts.quantile(q=0.75))

            x[self.columns_prefix + self.target_col + '_' + col + '_mean'] = mean_res
            x[self.columns_prefix + self.target_col + '_' + col + '_median'] = median_res
            x[self.columns_prefix + self.target_col + '_' + col + '_p25'] = p25
            x[self.columns_prefix + self.target_col + '_' + col + '_p75'] = p75

        x.drop(columns=['weekday', self.date_col, self.target_col], inplace=True)

        return x

    def transform(self, df):
        """target features"""
        check_is_fitted(self)

        assert isinstance(df, pd.DataFrame)
        assert isinstance(df, pd.DataFrame)

        x = self.weekday_groupby(df)
        df = pd.concat((df, x), axis=1)

        return df



import copy
import time
from collections import OrderedDict

import numpy as np
import torch
from torch import nn
from torch.utils.data import TensorDataset, DataLoader

from spinesUtils.asserts import ParameterValuesAssert, augmented_isinstance, ParameterTypeAssert
from spinesUtils.utils import Logger

from spinesTS.metrics import WMAPELoss, RMSELoss
from spinesTS.utils import seed_everything, check_is_fitted


logger = Logger(with_time=False)


@ParameterTypeAssert({
    'device': str
})
def detect_available_device(device='auto'):
    if device != 'auto':
        device = device.lower()
        auto_selector = False
    else:
        auto_selector = True

    tpu_available = False

    try:
        import torch_xla.core.xla_model as xm
        device = xm.xla_device()

        if device is not None:
            tpu_available = True

    except ImportError:
        if device == 'tpu':
            print("[ImportError]: torch_xla package is not installed."
                  "Consider run `python3 -m pip install torch_xla` in your terminal.")

    mps_available = False
    cuda_available = False
    cpu_available = False
    mps_use = False
    cuda_use = False
    cpu_use = False
    tpu_use = False

    if torch.backends.mps.is_available():
        mps_available = True

    if torch.cuda.is_available():
        cuda_available = True

    if torch.cpu.is_available():
        cpu_available = True

    if tpu_available and (auto_selector or device == 'tpu'):
        device = xm.xla_device()
        tpu_use = True
    elif cuda_available and (auto_selector or device == 'cuda'):
        device = 'cuda:0' if torch.cuda.device_count() > 1 else 'cuda'
        cuda_use = True
    elif mps_available and (auto_selector or device == 'mps'):
        device = 'mps'
        mps_use = True
    elif cpu_available and (auto_selector or device == 'cpu'):
        device = 'cpu'
        cpu_use = True

    blank_length = lambda s: ' ' * 3 if s is True else ' ' * 2
    string_format = f"MPS  available: {mps_available}{blank_length(mps_available)}| MPS  use: {mps_use}\n" \
                    f"CUDA available: {cuda_available}{blank_length(cuda_available)}| CUDA use: {cuda_use}\n" \
                    f"TPU  available: {tpu_available}{blank_length(tpu_available)}| TPU  use: {tpu_use}\n" \
                    f"CPU  available: {cpu_available}{blank_length(cpu_available)}| CPU  use: {cpu_use}"

    return device, string_format


def clear_torch_cache(device):
    if device == 'cuda':
        torch.cuda.empty_cache()
    elif device == 'mps':
        torch.mps.empty_cache()


@ParameterValuesAssert({
    'name': lambda s: augmented_isinstance(s, (None, str))
})
def get_loss_func(name=None):
    """get loss function

    Parameters
    ----------
    name: str, name of loss function, default None

    Returns
    -------
    object, loss function.
    """
    names = {'huber': nn.HuberLoss(), 'mse': nn.MSELoss(), 'mae': nn.L1Loss(),
             'wmape': WMAPELoss(), 'rmse': RMSELoss()}
    if isinstance(name, str):
        name = name.lower()
        return names[name]
    else:
        return names['mae']


class TorchModelMixin:
    """Provide pytorch models common mixin class.

    This class make it easy to write code like this:
        ```python
        class Model(TorchModelMixin):
            def __init__(self, *args, **kwargs):
                # need to set random seed if you needed
                # need to set device which to put your tensor, default to cuda/cuda:0 if your gpu is available, else to cpu
                super(Model, self).__init__(seed=None, device=None)

                self.model, self.loss_fn, self.optimizer = self.call()  # implement your model architecture

            def call(self):
                # model = your_model_class()
                # loss_fn = your_loss_function()
                # optimizer = your_optimizer_function()
                # return model, loss_fn, optimizer
                pass

            def fit(self, X, y, *args, **kwargs):
                return super().fit(X, y, *args, **kwargs)

            # def metric(self, y_true, y_pred):
                 # your metric, default to mae(mean absolute error)
                 # if you want to use other metrics, you need to override this function

        # To fit something
        model = Model(args, kwargs)
        model.fit(X, y)

        # To predict something
        y_pred = model.predict(X)
        ```

    Parameters
    ----------
    seed : int or None, random seed
    device : str or None, device name
    loss_fn: str or None, loss function

    Returns
    -------
    None

    """

    def __init__(self, seed=None, device='auto', loss_fn='mae') -> None:
        self.training_logs = {
            'time_cost': [],
            'epochs': [],
            'batches': [],
            'lrs': [],
            'train_loss': [],
            'train_accuracy': [],
            'test_loss': [],
            'test_accuracy': [],
            'current_p': []
        }

        seed_everything(seed)
        self.device, self.string_format = detect_available_device(device)
        self.loss_fn_name = loss_fn
        self.loss_fn = get_loss_func(loss_fn)
        self.current_patience = 0

        self.__spinesTS_is_fitted__ = False

    def call(self, *args, **kwargs):
        """To implement the model architecture.

        """
        raise NotImplementedError("To implement a spinesTS.nn model class, you must implement a call function.")

    def fit(self,
            X,
            y,
            epochs=3000,
            batch_size='auto',
            eval_set=None,
            loss_type='down',
            metrics_name='score',
            monitor='val_loss',
            min_delta=0,
            patience=10,
            lr_scheduler='CosineAnnealingLR',
            lr_scheduler_patience=10,
            lr_factor=0.1,
            restore_best_weights=True,
            verbose=True,
            **lr_scheduler_kwargs
            ):
        """Fit your model.

        Parameters
        ----------
        X : torch.Tensor, training features
        y : torch.Tensor, training targets
        epochs : int, training epochs, default to 1000
        batch_size : str or int, 'auto' means to autofit, int means to specify the batch size
        eval_set : iterable object(tuple or list) of torch.Tensor, default to None
        loss_type : str, 'down' or 'rise', only be used if lr_scheduler='ReduceLROnPlateau'
            it means the way to set the learning rate scheduler to watch the loss value (down or rise)
        metrics_name : str, names your metrics, default to 'score'
        monitor : str, 'val_loss' or 'loss', quantity to be monitored,
        min_delta : minimum change in the monitored quantity to qualify as an improvement,
            i.e. an absolute change of less than min_delta, will count as no improvement, default to 0
        patience : number of epochs with no improvement after which training will be stopped, default to 10
        lr_scheduler : learning rate scheduler name, one of ['ReduceLROnPlateau', 'CosineAnnealingLR',
            'CosineAnnealingWarmRestarts', None]
        lr_scheduler_patience :  number of epochs with no improvement after which learning rate will be reduced.
            For example, if patience = 2, then we will ignore the first 2 epochs with no improvement,
            and will only decrease the LR after the 3rd epoch  if the loss still hasn’t improved then, default: 10
        lr_factor : factor by which the learning rate will be reduced. new_lr = lr * factor. Default: 0.1
        restore_best_weights : Whether to restore model weights
                        from the epoch with the best value of the monitored quantity.
            If False, the model weights obtained at the last step of training are used.
            If True, and if no epoch improves, training will run for patience epochs and restore weights from
                the best epoch in that set. Default to True.
        verbose : Whether to  displays messages, default to True
        **lr_scheduler_kwargs : torch.optim.lr_scheduler parameters

        Returns
        -------
        self

        """
        if verbose:
            logger.print(self.string_format)
            time.sleep(0.5)

        return self._fit(
            X,
            y,
            epochs=epochs,
            batch_size=batch_size,
            eval_set=eval_set,
            loss_type=loss_type,
            metrics_name=metrics_name,
            monitor=monitor,
            min_delta=min_delta,
            patience=patience,
            lr_scheduler=lr_scheduler,
            lr_scheduler_patience=lr_scheduler_patience,
            lr_factor=lr_factor,
            restore_best_weights=restore_best_weights,
            verbose=verbose,
            **lr_scheduler_kwargs
        )

    def predict(self, X):
        """
        X : torch.Tensor, tensor which to predict
        """
        check_is_fitted(self)
        self.model.eval()
        with torch.no_grad():
            X = torch.Tensor(X)
            X = self._move_to_device(X)
            pred = self.model(X)
        return pred.cpu().numpy()

    def _move_to_device(self, obj):
        obj = obj.to(self.device)
        return obj

    def metric(self, y_true, y_pred):
        """model metric"""
        return nn.functional.l1_loss(y_true, y_pred).item()

    def _get_batch_size(self, x, batch_size='auto'):
        if batch_size == 'auto':
            self._batch_size = 2 ** np.log2(len(x)) * 16
            self._batch_size = 5096 if self._batch_size > 5096 else self._batch_size
        else:
            assert isinstance(batch_size, int) and batch_size > 0
            self._batch_size = batch_size

    @staticmethod
    def _check_x_y_type(X, y):
        if isinstance(X, np.ndarray):
            X = torch.from_numpy(X)
        elif not isinstance(X, torch.Tensor):
            X = torch.as_tensor(X)
        if isinstance(y, np.ndarray):
            y = torch.from_numpy(y)
        elif not isinstance(y, torch.Tensor):
            y = torch.as_tensor(y)

        return X.float(), y.float()

    def data_loader(self, X, y):
        train_data = TensorDataset(X, y)
        train_loader = DataLoader(train_data, batch_size=self._batch_size, shuffle=False)

        return train_loader

    def train_on_one_epoch(
            self,
            dataloader,
            model,
            loss_fn,
            optimizer
    ):
        """Training function on one epoch
        If you want to override it, you just need to return two values,
        current loss on this epoch, average-accuracy on this epoch
        """
        model.train()  # set model to training mode
        train_batch = len(dataloader)
        train_loss_current, train_metric = 0, 0

        for x, y in dataloader:
            clear_torch_cache(self.device)
            x, y = x.to(self.device), y.to(self.device)

            # compute error
            train_pred = model(x)
            train_loss = loss_fn(train_pred, y)
            optimizer.zero_grad(set_to_none=True)  # clear optimizer gradient

            # backward
            train_loss.backward()
            nn.utils.clip_grad_norm_(model.parameters(), max_norm=10.)

            optimizer.step()

            train_metric += self.metric(train_pred, y)
            train_loss_current += train_loss.item()  # 释放图形指针

        return train_loss_current / train_batch, train_metric / train_batch

    def test_on_one_epoch(
            self,
            dataloader,
            model,
            loss_fn
    ):
        """
        Test function on one epoch
        If you want to override it, you just need to return two values,
        current loss on this epoch, average-accuracy on this epoch

        """

        model.eval()  # set model to evaluate mode
        test_loss, test_metric, test_num_batches = 0, 0, len(dataloader)
        with torch.no_grad():  # with no gradient
            for x, y in dataloader:
                clear_torch_cache(self.device)
                x, y = x.to(self.device), y.to(self.device)
                pred = model(x)
                test_loss += loss_fn(y, pred).item()  # scalar
                # scalar
                test_metric += self.metric(y, pred)

        test_loss /= test_num_batches
        test_metric /= test_num_batches

        return test_loss, test_metric

    @ParameterTypeAssert({
        'loss_type': str,
        'min_delta': int,
        'patience': int,
        'restore_best_weights': bool
    })
    @ParameterValuesAssert({
        'loss_type': ('min', 'max')
    })
    def _early_stopping(
            self,
            loss,
            loss_type='min',
            min_delta=0,
            patience=10,
            restore_best_weights=True
    ):
        """
        loss type : min or max
        """
        if loss_type == 'max':
            loss = -loss

        if loss < (self.current_loss + min_delta):
            self.current_loss = loss
            if restore_best_weights:
                self.best_weight = copy.deepcopy(self.model.state_dict())
            self.current_patience = 0
        else:
            self.current_patience += 1

        if self.current_patience == patience:
            if restore_best_weights:
                self.model.load_state_dict(self.best_weight)
            return True

        return False

    def _get_lr_scheduler(
            self, mode=None,
            lr_scheduler='CosineAnnealingLR',
            lr_scheduler_patience=10,
            lr_factor=0.1,
            **lr_scheduler_kwargs
    ):
        if lr_scheduler == 'ReduceLROnPlateau':
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                self.optimizer,
                mode=mode,
                patience=lr_scheduler_patience, factor=lr_factor, **lr_scheduler_kwargs
            )
        elif lr_scheduler == 'CosineAnnealingLR':
            scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(self.optimizer, T_max=5, eta_min=0,
                                                                   **lr_scheduler_kwargs)
        elif lr_scheduler == 'CosineAnnealingWarmRestarts':
            scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(self.optimizer, T_0=5, T_mult=2,
                                                                             **lr_scheduler_kwargs)
        elif lr_scheduler is None:
            return None
        else:
            raise KeyError(f"{lr_scheduler} is invalid.")

        return scheduler

    def print_training_log(self, init_lr, metrics_name, total_epochs):
        training_log_print = OrderedDict({
            "epoch_msg": "",
            "current_p_msg": "",
            "lr_msg": "",
            "loss_msg": "",
            "val_loss_msg": "",
            "time_msg": ""
        })
        training_log_print["epoch_msg"] = f"Epoch {self.training_logs['epochs'][-1] + 1:>1d}/" + \
                                          f"{total_epochs:>1d}  " + \
                                          f"\n\r{self.training_logs['batches'][-1]}/{self.training_logs['batches'][-1]}"

        training_log_print["lr_msg"] = f"[*lr: {self.training_logs['lrs'][-1]:4e}]" if len(
            str(self.training_logs['lrs'][-1])) <= 5 else f" [*lr: {self.training_logs['lrs'][-1]:>5}]"

        training_log_print["loss_msg"] = f"loss: {self.training_logs['train_loss'][-1]:>.4f} - " + \
                                         f"{metrics_name}: {self.training_logs['train_accuracy'][-1]:>.4f}"

        training_log_print[
            "val_loss_msg"
        ] = f"val_loss: {self.training_logs['test_loss'][-1]:>.4f} - " + \
            f"val_{metrics_name}: {self.training_logs['test_accuracy'][-1]:>.4f}"

        training_log_print["time_msg"] = \
            f"{self.training_logs['time_cost'][-1]:>.2f}s/epoch - " + \
            f"{self.training_logs['time_cost'][-1] / self.training_logs['batches'][-1]:>.3f}s/step"

        training_log_print["current_p_msg"] = f"p{self.training_logs['current_p'][-1]}"

        if round(self.training_logs['lrs'][-1], 5) == init_lr:
            del training_log_print['lr_msg']
        metric_string = ' - '.join([i for i in training_log_print.values()])

        return metric_string

    @staticmethod
    def _check_eval_set_params(eval_set):
        if isinstance(eval_set, tuple):
            assert len(eval_set) == 2
        elif isinstance(eval_set, list):
            assert len(eval_set[0]) == 2
            eval_set = eval_set[0]

        return eval_set

    def _fit(
            self,
            X,
            y,
            epochs=1000,
            batch_size='auto',
            eval_set=None,
            loss_type='min',
            metrics_name='score',
            monitor='val_loss',
            min_delta=0,
            patience=10,
            lr_scheduler='CosineAnnealingLR',
            lr_scheduler_patience=10,
            lr_factor=0.1,
            restore_best_weights=True,
            verbose=True,
            **lr_scheduler_kwargs
    ):
        """
        lr_scheduler: torch.optim.lr_scheduler class,
            only support to ['ReduceLROnPlateau', 'CosineAnnealingLR', 'CosineAnnealingWarmRestarts']
        """

        assert eval_set is None or isinstance(eval_set, (list, tuple))
        assert monitor in ('train_loss', 'val_loss', None)

        init_lr = copy.deepcopy(self.learning_rate)

        self.model = self._move_to_device(self.model)
        X, y = self._check_x_y_type(X, y)

        eval_set = self._check_eval_set_params(eval_set)

        self._get_batch_size(X, batch_size=batch_size)
        train_dataloader = self.data_loader(X, y)
        test_dataloader = None

        if eval_set is not None:
            test_dataloader = self.data_loader(*self._check_x_y_type(eval_set[0], eval_set[1]))

        self.current_loss = np.finfo(np.float64).max - min_delta
        self.best_weight = copy.deepcopy(self.model.state_dict())

        batches = int(np.ceil(len(X) / self._batch_size))

        mode = loss_type

        scheduler = self._get_lr_scheduler(
            mode=mode, lr_scheduler=lr_scheduler,
            lr_scheduler_patience=lr_scheduler_patience,
            lr_factor=lr_factor,
            **lr_scheduler_kwargs)

        for epoch in range(epochs):
            tik = time.time()
            stop_state = False

            self.training_logs['epochs'].append(epoch)
            self.training_logs['batches'].append(batches)

            train_loss_current, train_acc = self.train_on_one_epoch(train_dataloader, model=self.model,
                                                                    loss_fn=self.loss_fn,
                                                                    optimizer=self.optimizer)

            if lr_scheduler:
                scheduler.step() if lr_scheduler != 'ReduceLROnPlateau' else scheduler.step(train_loss_current)

            self.training_logs['lrs'].append(round(float(self.optimizer.state_dict()['param_groups'][0]['lr']), 7))

            self.training_logs['train_loss'].append(train_loss_current)
            self.training_logs['train_accuracy'].append(train_acc)

            if monitor == 'train_loss':
                stop_state = self._early_stopping(train_loss_current, loss_type=loss_type,
                                                  min_delta=min_delta, patience=patience,
                                                  restore_best_weights=restore_best_weights)
            else:
                if test_dataloader:
                    test_loss, test_acc = self.test_on_one_epoch(test_dataloader, self.model, self.loss_fn)
                    stop_state = self._early_stopping(test_loss, loss_type=loss_type, min_delta=min_delta,
                                                      patience=patience, restore_best_weights=restore_best_weights)

                    self.training_logs['test_loss'].append(test_loss)
                    self.training_logs['test_accuracy'].append(test_acc)

            if monitor is not None and eval_set is not None:
                self.training_logs['current_p'].append(self.current_patience)

            tok = time.time()

            self.training_logs['time_cost'].append(tok - tik)

            if verbose:
                print(self.print_training_log(init_lr, metrics_name, epochs))

            if stop_state:
                if verbose:
                    print(f"Early stopping at epoch {epoch}.")
                break

        self.__spinesTS_is_fitted__ = True
        return self

    def score(self, X, y):
        self.model.eval()
        with torch.no_grad():
            X, y = torch.Tensor(X), torch.Tensor(y)
            X_gpu, _ = self._move_to_device(X), self._move_to_device(y)
            pred = self.model(X_gpu).cpu().numpy()

        return self.metric(y, pred)



import numpy as np
from joblib import Parallel, delayed
from scipy.interpolate import interp1d
from scipy.special import erf, erfinv
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import FLOAT_DTYPES, check_array, check_is_fitted
import torch
from sklearn.preprocessing import MinMaxScaler


class GaussRankScaler(BaseEstimator, TransformerMixin):
    """Transform features by scaling each feature to a normal distribution.

    Parameters
    -----------
    epsilon : float, optional, default 1e-4
        A small amount added to the lower bound or subtracted
        from the upper bound. This value prevents infinite number
        from occurring when applying the inverse error function.
    copy : boolean, optional, default True
        If False, try to avoid a copy and do inplace scaling instead.
        This is not guaranteed to always work inplace; e.g. if the data is
        not a NumPy array, a copy may still be returned.
    n_jobs : int or None, optional, default None
        Number of jobs to run in parallel.
        ``None`` means 1 and ``-1`` means using all processors.
    interp_kind : str or int, optional, default 'linear'
       Specifies the kind of interpolation as a string
        ('linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic',
        'previous', 'next', where 'zero', 'slinear', 'quadratic' and 'cubic'
        refer to a spline interpolation of zeroth, first, second or third
        order; 'previous' and 'next' simply return the previous or next value
        of the point) or as an integer specifying the order of the spline
        interpolator to use.
    interp_copy : bool, optional, default False
        If True, the interpolation function makes internal copies of x and y.
        If False, references to `x` and `y` are used.

    Returns
    -------
    self, object
    """

    def __init__(self, epsilon=1e-4, copy=True, n_jobs=None, interp_kind='linear', interp_copy=False):
        self.epsilon = epsilon
        self.copy = copy
        self.interp_kind = interp_kind
        self.interp_copy = interp_copy
        self.fill_value = 'extrapolate'
        self.n_jobs = n_jobs

    def fit(self, X, y=None):
        """Fit interpolation function to link rank with original data for future scaling

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The data used to fit interpolation function for later scaling along the features axis.
        y
            Ignored

        Returns
        -------
        self, object

        """
        X = check_array(X, copy=self.copy, estimator=self, dtype=FLOAT_DTYPES, force_all_finite=True)

        self.interp_func_ = Parallel(n_jobs=self.n_jobs)(delayed(self._fit)(x) for x in X.T)
        return self

    def _fit(self, x):
        x = self.drop_duplicates(x)
        rank = np.argsort(np.argsort(x))
        bound = 1.0 - self.epsilon
        factor = np.max(rank) / 2.0 * bound
        scaled_rank = np.clip(rank / factor - bound, -bound, bound)
        return interp1d(
            x, scaled_rank, kind=self.interp_kind, copy=self.interp_copy, fill_value=self.fill_value)

    def transform(self, X, copy=None):
        """Scale the data with the Gauss Rank algorithm

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The data used to scale along the features axis.
        copy : bool, optional (default: None)
            Copy the input X or not.

        Returns
        -------
        numpy.ndarray
        """
        check_is_fitted(self, 'interp_func_')

        copy = copy if copy is not None else self.copy
        X = check_array(X, copy=copy, estimator=self, dtype=FLOAT_DTYPES, force_all_finite=True)

        X = np.array(Parallel(n_jobs=self.n_jobs)(delayed(self._transform)(i, x) for i, x in enumerate(X.T))).T
        return X

    def _transform(self, i, x):
        return erfinv(self.interp_func_[i](x))

    def inverse_transform(self, X, copy=None):
        """Scale back the data to the original representation

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data used to scale along the features axis.
        copy : bool, optional (default: None)
            Copy the input X or not.

        Returns
        -------
        numpy.ndaray

        """
        check_is_fitted(self, 'interp_func_')

        copy = copy if copy is not None else self.copy
        X = check_array(X, copy=copy, estimator=self, dtype=FLOAT_DTYPES, force_all_finite=True)

        X = np.array(Parallel(n_jobs=self.n_jobs)(delayed(self._inverse_transform)(i, x) for i, x in enumerate(X.T))).T
        return X

    def _inverse_transform(self, i, x):
        inv_interp_func = interp1d(self.interp_func_[i].y, self.interp_func_[i].x, kind=self.interp_kind,
                                   copy=self.interp_copy, fill_value=self.fill_value)
        return inv_interp_func(erf(x))

    @staticmethod
    def drop_duplicates(x):
        is_unique = np.zeros_like(x, dtype=bool)
        is_unique[np.unique(x, return_index=True)[1]] = True
        return x[is_unique]


class MultiDimScaler:
    def __init__(self):
        self.scalers = []

    def fit(self, x, scaler=MinMaxScaler):
        assert x.ndim == 3

        for i in range(x.shape[-1]):
            self.scalers.append(scaler().fit(x[:, :, i]))

        return self

    def transform(self, x):
        assert len(self.scalers) == x.shape[-1], "MultiDimScaler not fitted yet."

        for i in range(x.shape[-1]):
            x[:, :, i] = self.scalers[i].transform(x[:, :, i])

        if isinstance(x, np.ndarray):
            x = torch.from_numpy(x)
        elif not isinstance(x, torch.Tensor):
            x = torch.Tensor(x)

        return x.float()

    def fit_transform(self, x, scaler=MinMaxScaler):
        return self.fit(x, scaler).transform(x)

    def inverse_transform(self, x):
        assert len(self.scalers) == x.shape[-1], "MultiDimScaler not fitted yet."

        if isinstance(x, torch.Tensor):
            x = x.numpy()

        for i in range(x.shape[-1]):
            x[:, :, i] = self.scalers[i].inverse_transform(x[:, :, i])

        return torch.Tensor(x).float()



from ._data_generator import DataGenerator, RandomEventGenerator

from ._data_reader import (
    LoadElectricDataSets,
    LoadMessagesSentHourDataSets,
    LoadMessagesSentDataSets,
    LoadWebSales,
    LoadSupermarketIncoming,
    BuiltInSeriesData
)





import copy
import warnings

import numpy as np
import torch
from joblib import Parallel, delayed
from sklearn.multioutput import _MultiOutputEstimator, _fit_estimator
from sklearn.utils.validation import _check_fit_params

from spinesTS.base import ForecastingMixin
from spinesTS.utils import func_has_params, check_is_fitted

warnings.filterwarnings('ignore')


class MultiStepRegressor(ForecastingMixin):
    """Use the last predict-step value as the last step true value,
        to predict current step value,
        and repeat this until it reaches y.shape[1] times.

    Parameters
    ----------
    estimator: model with implement fit and predict function

    Returns
    -------
    self : object
    """

    def __init__(self, estimator):
        self._estimator = estimator
        self._forward = 1
        self.__spinesTS_is_fitted__ = False

    def fit(self, x, y, eval_set=None, **estimator_fit_kwargs):
        assert np.ndim(y) <= 2
        if eval_set is not None:
            if len(eval_set) == 2 or len(eval_set[0]) == 2:
                if len(eval_set[0]) == 2:
                    eval_set = eval_set[0]
            else:
                raise ValueError("The eval_set must be [X, y] or [(X, y)]")

        self._forward = y.shape[1]
        _ = y
        if np.ndim(y) == 2:
            assert np.ndim(y) <= 2
            _ = y[:, 0]

        if func_has_params(self._estimator.fit, "eval_set") and eval_set is not None:

            try:
                eval_sets_ = (eval_set[0], eval_set[1][:, 0])
                self._estimator.fit(x, _, eval_set=eval_sets_, **estimator_fit_kwargs)
            except Exception:
                eval_sets_ = [(eval_set[0], eval_set[1][:, 0])]
                self._estimator.fit(x, _, eval_set=eval_sets_, **estimator_fit_kwargs)
        else:
            self._estimator.fit(x, _, **estimator_fit_kwargs)

        self.__spinesTS_is_fitted__ = True
        return self

    def predict(self, x, **kwargs):
        check_is_fitted(self)
        assert isinstance(x, np.ndarray)
        res = []
        eval_x = copy.deepcopy(x)
        for step in range(self._forward):  # forward n steps
            _ = np.squeeze(self._estimator.predict(eval_x, **kwargs))
            res.append(_)
            eval_x = np.concatenate([eval_x[:, 1:], np.transpose(_.reshape(1, -1))], axis=1)
            assert x.shape == eval_x.shape

        r = np.asarray(res)

        return np.squeeze(np.transpose(r))


class MultiOutputRegressor(ForecastingMixin, _MultiOutputEstimator):
    """Fitting one regressor per target.

    Parameters
    ----------
    estimator: model with implement fit and predict function
    n_jobs: None or int, default to None.
        The number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
        context. ``-1`` means using all processors.

    Returns
    -------
    self : object

    """

    def __init__(self, estimator, n_jobs=None):
        super().__init__(estimator, n_jobs=n_jobs)

    def fit(self, X, y, eval_set=None, **fit_params):
        """Fit the model to data, separately for each output variable.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input data.
        y : {array-like, sparse matrix} of shape (n_samples, n_outputs)
            Multi-output targets. An indicator matrix turns on multilabel
            estimation.
        eval_set : tuple of (X, y) , length must be equal to y.shape[1],
            passed to the ``estimator.fit`` method of each step
        **fit_params : dict of string -> object
            Parameters passed to the ``estimator.fit`` method of each step.

        Returns
        -------
        self : object
            Returns a fitted instance.
        """
        if eval_set is not None:
            if len(eval_set) == 2 or len(eval_set[0]) == 2:
                if len(eval_set[0]) == 2:
                    eval_set = eval_set[0]
            else:
                raise ValueError("The eval_set must be [X, y] or [(X, y)]")

        if not hasattr(self.estimator, "fit"):
            raise ValueError("The base estimator should implement a fit method")

        y = self._validate_data(X="no_validation", y=y, multi_output=True)

        if y.ndim == 1:
            raise ValueError(
                "y must have at least two dimensions for "
                "multi-output regression but has only one."
            )

        fit_params_validated = _check_fit_params(X, fit_params)

        if func_has_params(self.estimator.fit, "eval_set") and eval_set is not None:
            try:
                eval_sets_ = [(eval_set[0], eval_set[1][:, i]) for i in range(y.shape[1])]
                self.estimators_ = Parallel(n_jobs=self.n_jobs)(
                    delayed(_fit_estimator)(
                        self.estimator, X, y[:, i], eval_set=eval_sets_[i], **fit_params_validated
                    )
                    for i in range(y.shape[1])
                )
            except Exception:
                eval_sets_ = [[(eval_set[0], eval_set[1][:, i])] for i in range(y.shape[1])]
                self.estimators_ = Parallel(n_jobs=self.n_jobs)(
                    delayed(_fit_estimator)(
                        self.estimator, X, y[:, i], eval_set=eval_sets_[i], **fit_params_validated
                    )
                    for i in range(y.shape[1])
                )
        else:
            self.estimators_ = Parallel(n_jobs=self.n_jobs)(
                delayed(_fit_estimator)(
                    self.estimator, X, y[:, i], **fit_params_validated
                )
                for i in range(y.shape[1])
            )

        if hasattr(self.estimators_[0], "n_features_in_"):
            self.n_features_in_ = self.estimators_[0].n_features_in_
        if hasattr(self.estimators_[0], "feature_names_in_"):
            self.feature_names_in_ = self.estimators_[0].feature_names_in_

        return self


class MultivariableMultiOutputRegressor(ForecastingMixin):
    """多变量多输出预测"""

    def __init__(self):
        pass

    def check_x_y(self, x, y):
        assert isinstance(x, (np.ndarray, torch.Tensor)) and isinstance(y, (np.ndarray, torch.Tensor)), \
            "only accept numpy.ndarray or torch.Tensor"
        assert x.ndim == 3

    def fit(self, x, **kwargs):
        pass

    def predict(self, x, n, **kwargs):
        pass



import numpy as np
from sklearn.metrics import *
from torch import nn
import torch


def wmape(y_true, y_pred):
    return np.abs(y_true - y_pred).sum() / np.abs(y_true).sum()


def rmse(*args, **kwargs):
    return mean_squared_error(*args, **kwargs) ** 0.5


def mae(*args, **kwargs):
    return mean_absolute_error(*args, **kwargs)


def mse(*args, **kwargs):
    return mean_squared_error(*args, **kwargs)


class WMAPELoss(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(WMAPELoss, self).__init__()

    def forward(self, inputs, targets):
        return torch.abs(inputs - targets).sum() / torch.abs(inputs).sum()


class RMSELoss(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(RMSELoss, self).__init__()
        self.mse = nn.MSELoss()

    def forward(self, inputs, targets):
        return torch.sqrt(self.mse(inputs, targets))


from ._multistep_forecast import MultiStepRegressor, MultiOutputRegressor
from ._io_model import load_model, save_model
from ._gbrt import GBRTPreprocessing, WideGBRT
from ._dummy_model import DummyModel



from typing import Any, Union

import torch
from torch import nn

from spinesUtils.preprocessing import unsqueeze_if

from spinesTS.layers import GAU, PositionalEncoding
from spinesTS.base import TorchModelMixin, ForecastingMixin


class GAUBlock(nn.Module):
    def __init__(self, in_features, level=2, **kwargs):
        super(GAUBlock, self).__init__()
        self.gau = nn.ModuleList([
            nn.Sequential(
                PositionalEncoding(in_features, add_x=True),
                GAU(in_features, **kwargs)
            )
            for i in range(level)
            ])
        self.level = level

    def forward(self, x):
        x = unsqueeze_if(x, x.ndim == 2, 1)
        for i in self.gau:
            x = i(x)

        return x


class GAUBase(nn.Module):
    def __init__(self, in_shapes, out_features, level=2):
        super(GAUBase, self).__init__()
        self.in_shapes_type = type(in_shapes)

        self.in_features, self.out_features = \
            in_shapes[-1] if self.in_shapes_type == tuple else in_shapes, out_features

        self.gau = GAUBlock(self.in_features, level)

        ln_layer_in_fea = in_shapes[0] * in_shapes[1] if self.in_shapes_type == tuple else self.in_features

        self.linear = nn.Linear(ln_layer_in_fea, out_features)

    def forward(self, x):
        x = self.gau(x)

        if x.ndim == 2:
            return self.linear(x)

        return self.linear(x.reshape(-1, x.shape[1] * x.shape[2]))


class GAUNet(TorchModelMixin, ForecastingMixin):
    def __init__(self,
                 in_features: Any,
                 out_features: Any,
                 level: int = 2,
                 learning_rate: float = 0.01,
                 random_seed: int = 42,
                 device='auto',
                 loss_fn='mae'
                 ) -> None:
        super(GAUNet, self).__init__(random_seed, device, loss_fn=loss_fn)
        self.in_features, self.out_features = in_features, out_features
        self.learning_rate = learning_rate
        self.loss_fn_name = loss_fn
        self.model, self.loss_fn, self.optimizer = self.call(level)

    def call(self,
             level: int = 2) -> tuple:
        model = GAUBase(self.in_features, self.out_features,
                        level=level)
        loss_fn = self.loss_fn
        optimizer = torch.optim.AdamW(model.parameters(), lr=self.learning_rate)
        return model, loss_fn, optimizer

    def fit(self,
            X_train: Any,
            y_train: Any,
            epochs: int = 1000,
            batch_size: Union[str, int] = 'auto',
            eval_set: Any = None,
            monitor: str = 'val_loss',
            min_delta: int = 0,
            patience: int = 10,
            lr_scheduler: Union[str, None] = 'CosineAnnealingLR',
            lr_scheduler_patience: int = 10,
            lr_factor: float = 0.7,
            restore_best_weights: bool = True,
            verbose: bool = True,
            loss_type='min',
            **kwargs: Any) -> Any:
        return super().fit(X_train, y_train, epochs, batch_size, eval_set, loss_type=loss_type,
                           metrics_name=self.loss_fn_name,
                           monitor=monitor, lr_scheduler=lr_scheduler,
                           lr_scheduler_patience=lr_scheduler_patience,
                           lr_factor=lr_factor,
                           min_delta=min_delta, patience=patience, restore_best_weights=restore_best_weights,
                           verbose=verbose, **kwargs)



from typing import Any, Union

import torch
from torch import nn

from spinesTS.base import TorchModelMixin, ForecastingMixin
from spinesTS.layers import PositionalEncoding


class EncoderDecoderBlock(nn.Module):
    def __init__(self, in_features, out_features, bias=False,
                 dropout=0.):
        super(EncoderDecoderBlock, self).__init__()

        self.encoder = nn.LSTM(in_features, in_features, num_layers=1,
                               bias=bias, dropout=dropout,
                               bidirectional=False, batch_first=True)

        self.decoder_output = nn.Linear(in_features, in_features)

        self.position_encoder = PositionalEncoding(in_features)

        self.linear = nn.Linear(in_features, out_features)
        self.gelu = nn.GELU()

    def forward(self, x):
        if x.ndim == 2:
            x = torch.unsqueeze(x, dim=1)

        _, (h, _) = self.encoder(x)

        x = self.position_encoder(x)
        x = x.squeeze() + h.squeeze()

        output = self.decoder_output(x.squeeze())

        output = self.gelu(self.linear(output.squeeze()))

        return output


class Seq2SeqBlock(nn.Module):
    def __init__(self, in_features, out_features, bias=False,
                 dropout=0.,):
        super(Seq2SeqBlock, self).__init__()

        self.in_features, self.out_features = in_features, out_features

        self.forward_block = EncoderDecoderBlock(
                    in_features=in_features,
                    out_features=in_features,
                    bias=bias, dropout=dropout
        )

        self.backward_block = EncoderDecoderBlock(
            in_features=in_features,
            out_features=in_features,
            bias=bias, dropout=dropout
        )

        self.linear_0 = nn.Linear(in_features * 2, 256)
        self.gelu = nn.GELU()
        self.linear_1 = nn.Linear(256, out_features)

        self.position_encoder1 = PositionalEncoding(in_features, add_x=False)
        self.position_encoder2 = PositionalEncoding(in_features, add_x=False)

    def forward(self, x):
        backward_output = self.backward_block(torch.flip(x, dims=[-1]))
        forward_output = self.forward_block(x)

        if backward_output.ndim == 1:
            backward_output = backward_output.unsqueeze(0)
        if forward_output.ndim == 1:
            forward_output = forward_output.unsqueeze(0)

        forward_output = forward_output + self.position_encoder1(forward_output).squeeze().mean(dim=1)
        backward_output = backward_output + self.position_encoder2(backward_output).squeeze().mean(dim=1)

        output = torch.concat((forward_output, backward_output), dim=-1)
        output = self.linear_0(output.squeeze())
        output = self.gelu(output)

        return self.linear_1(output)


class StackingRNN(TorchModelMixin, ForecastingMixin):
    def __init__(self,
                 in_features: int,
                 out_features: int,
                 loss_fn='mae',
                 bias=False,
                 dropout=0.1,
                 learning_rate: float = 0.001,
                 random_seed: int = 42,
                 device='auto'
                 ) -> None:
        super(StackingRNN, self).__init__(random_seed, device, loss_fn=loss_fn)
        self.in_features, self.out_features = in_features, out_features
        self.learning_rate = learning_rate
        self.model, self.loss_fn, self.optimizer = self.call(bias=bias,
                                                             dropout=dropout)
        self.loss_fn_name = loss_fn

    def call(self, bias,
             dropout) -> tuple:
        model = Seq2SeqBlock(
            in_features=self.in_features,
            out_features=self.out_features,
            bias=bias,
            dropout=dropout,
        )
        loss_fn = self.loss_fn
        optimizer = torch.optim.AdamW(model.parameters(), lr=self.learning_rate)
        return model, loss_fn, optimizer

    def fit(self,
            X_train: Any,
            y_train: Any,
            epochs: int = 1000,
            batch_size: Union[str, int] = 'auto',
            eval_set: Any = None,
            monitor: str = 'val_loss',
            min_delta: int = 0,
            patience: int = 10,
            lr_scheduler: Union[str, None] = 'CosineAnnealingLR',
            lr_scheduler_patience: int = 10,
            lr_factor: float = 0.7,
            restore_best_weights: bool = True,
            verbose: bool = True,
            loss_type='min',
            **kwargs: Any) -> Any:
        return super().fit(X_train, y_train, epochs, batch_size, eval_set, loss_type=loss_type,
                           metrics_name=self.loss_fn_name,
                           monitor=monitor, lr_scheduler=lr_scheduler,
                           lr_scheduler_patience=lr_scheduler_patience,
                           lr_factor=lr_factor,
                           min_delta=min_delta, patience=patience, restore_best_weights=restore_best_weights,
                           verbose=verbose, **kwargs)



from io import StringIO

import pandas as pd
from pandas import get_option
from pandas.io.formats import format as fmt


class DataTS(pd.DataFrame):
    """Convert a pandas dataframe to a time series dataframe.

    """
    def __init__(self, dataset, name=None, date_column=None, format=None, unit=None):
        super().__init__(data=dataset)
        self.date_column = date_column
        self.format = format
        self.dataset_name = name or 'spinesTS.DataTS'
        if date_column is not None:
            self[date_column] = pd.to_datetime(self[date_column], format=format, unit=unit)

            self.convert2datetime()

    def convert2datetime(self):
        # 将传入的pandas dataframe转换为以时间为索引的dataframe
        self.set_index(self.date_column, inplace=True, drop=True)

    def __str__(self):
        return f"{self.dataset_name} dataset, shape={self.shape}, head_data=\n{self.head()}"

    def _repr__(self):
        return self.__str__()

    def _repr_html_(self) -> str | None:
        """Return a html representation for a particular DataFrame.

        Mainly for IPython notebook.
        """
        if self._info_repr():
            buf = StringIO()
            self.info(buf=buf)
            # need to escape the <class>, should be the first line.
            val = buf.getvalue().replace("<", r"&lt;", 1)
            val = val.replace(">", r"&gt;", 1)
            return f"<pre>{val}</pre>"

        if get_option("display.notebook_repr_html"):
            min_rows = get_option("display.min_rows")
            max_cols = get_option("display.max_columns")
            show_dimensions = get_option("display.show_dimensions")

            formatter = fmt.DataFrameFormatter(
                self.head(),
                columns=None,
                col_space=None,
                na_rep="NaN",
                formatters=None,
                float_format=None,
                sparsify=None,
                justify=None,
                index_names=True,
                header=True,
                index=True,
                bold_rows=True,
                escape=True,
                max_rows=5,
                min_rows=min_rows,
                max_cols=max_cols,
                show_dimensions=show_dimensions,
                decimal=".",
            )

            prefix = f"<div><b>{self.dataset_name}</b>"
            postfix = f"shape: {self.shape}</div>"
            return prefix + fmt.DataFrameRenderer(formatter).to_html(notebook=True) + postfix
        else:
            return None
